import datetime
from collections import namedtuple

from django.utils.encoding import force_text
from django.db.backends.base.introspection import (
    BaseDatabaseIntrospection, FieldInfo, TableInfo,
)

FieldInfo = namedtuple('FieldInfo', FieldInfo._fields + ('identity_type',))
InfoLine = namedtuple('InfoLine', 'col_name data_type max_len num_prec num_scale extra column_default identity_type')


class DatabaseIntrospection(BaseDatabaseIntrospection):
    # Maps type codes to Django Field types.
    data_types_reverse = {
        23: 'BooleanField',
        7: 'SmallIntegerField',
        8: 'IntegerField',
        10: 'FloatField',
        12: 'DateField',
        13: 'TimeField',
        14: 'CharField',
        16: 'BigIntegerField',
        27: 'FloatField',
        35: 'DateTimeField',
        37: 'CharField',
        40: 'TextField',
        260: 'BinaryField',
        261: 'TextField',
        # A NUMERIC/DECIMAL data type is stored as a SMALLINT, INTEGER or BIGINT
        # in Firebird, thus the value of RDB$FIELD_TYPE is reported. So we need
        # two additional helper data types for that to distinguish between real
        # Integer data types and NUMERIC/DECIMAL
        161: 'DecimalField',  # NUMERIC => RDB$FIELD_SUB_TYPE = 1
        162: 'DecimalField',  # DECIMAL => RDB$FIELD_SUB_TYPE = 2
        # Also, the scale value of a NUMERIC/DECIMAL fields is stored as negative
        # number in the Firebird system tables, thus we have to multiply with -1.
        # The SELECT statement in the function get_table_description takes care
        # of all of that.
    }

    def table_name_converter(self, name):
        return name.lower()

    def sequence_list(self):
        with self.connection.cursor() as cursor:
            cursor.execute(
                """SELECT lower(trim(rdb$relation_name)), lower(trim(rdb$field_name))
                    FROM rdb$relation_fields
                    WHERE rdb$identity_type is not null
                    """)
            return [{'table': r[0], 'column': r[1]} for r in cursor.fetchall()]

    def get_field_type(self, data_type, description):
        field_type = super().get_field_type(data_type, description)
        if description.identity_type:
            if field_type == 'IntegerField':
                return 'AutoField'
            elif field_type == 'BigIntegerField':
                return 'BigAutoField'
        return field_type

    def get_table_list(self, cursor):
        "Returns a list of table names in the current database."
        cursor.execute("""
            select
                lower(trim(rdb$relation_name)),
                case when RDB$VIEW_BLR IS NULL then 't' else 'v' end as rel_type
            from rdb$relations
            where rdb$system_flag=0
            order by 1 """)
        # return [r[0].strip().lower() for r in cursor.fetchall()]
        return [TableInfo(row[0], row[1]) for row in cursor.fetchall()]

    def get_table_description(self, cursor, table_name):
        """
        Returns a description of the table, with the DB-API cursor.description interface.
        Must return a 'FieldInfo' struct 'name type_code display_size internal_size precision scale null_ok'
        """
        tbl_name = "'%s'" % table_name.upper()
        cursor.execute("""
            select
              lower(trim(rf.rdb$field_name))
              , case
                  when (f.rdb$field_type in (7,8,16)) and (f.rdb$field_sub_type > 0) then
                    160 + f.rdb$field_sub_type
                  when (f.rdb$field_type = 261) then
                    260 + f.rdb$field_sub_type
                  else
                    f.rdb$field_type
                end as type_code

              , case
                  when (f.rdb$field_type in (14,37)) then
                    f.rdb$character_length
                  else
                    f.rdb$field_length
                end as field_length
              , f.rdb$field_precision
              , f.rdb$field_scale * -1
              , rf.rdb$null_flag
              , rf.rdb$default_source
              , rf.rdb$identity_type
            from
              rdb$relation_fields rf join rdb$fields f on (rf.rdb$field_source = f.rdb$field_name)
            where
              upper(rf.rdb$relation_name) = %s
            order by
              rf.rdb$field_position
            """ % (tbl_name,))
        items = []
        for r in cursor.fetchall():
            # name type_code display_size internal_size precision scale null_ok, default, identity_type
            items.append(FieldInfo(r[0], r[1], r[2], r[2] or 0, r[3], r[4], not (r[5] == 1), r[6], r[7]))
        return items

    def _name_to_index(self, cursor, table_name):
        """Return a dictionary of {field_name: field_index} for the given table.
           Indexes are 0-based.
        """
        return dict([(d[0], i) for i, d in enumerate(self.get_table_description(cursor, table_name))])

    def _get_references(self, table_name):
        """
        Foreign Key constraint reference.
        constraint name, table, column, referenecd table, referenced column
        table_name: refernced tbale name
        """
        with self.connection.cursor() as cursor:
            tbl_name = "'%s'" % table_name.upper()
            references = []
            cursor.execute("""
                select
                    refc.rdb$constraint_name,
                    i.rdb$relation_name
                from rdb$index_segments s
                left join rdb$indices i on i.rdb$index_name = s.rdb$index_name
                left join rdb$relation_constraints rc on rc.rdb$index_name = s.rdb$index_name
                left join rdb$ref_constraints refc on rc.rdb$constraint_name = refc.rdb$constraint_name
                left join rdb$relation_constraints rc2 on rc2.rdb$constraint_name = refc.rdb$const_name_uq
                left join rdb$indices i2 on i2.rdb$index_name = rc2.rdb$index_name
                left join rdb$index_segments s2 on i2.rdb$index_name = s2.rdb$index_name
                WHERE RC.RDB$CONSTRAINT_TYPE = 'FOREIGN KEY'
                and upper(i2.rdb$relation_name) = %s """ % (tbl_name,))
            return [(r[0].strip(), r[1].strip()) for r in cursor.fetchall()]

    def get_key_columns(self, cursor, table_name):
        """
        Backends can override this to return a list of (column_name, referenced_table_name,
        referenced_column_name) for all key columns in given table.
        """
        tbl_name = "'%s'" % table_name.upper()
        key_columns = []
        cursor.execute("""
            select
                lower(s.rdb$field_name) as column_name,
                lower(i2.rdb$relation_name) as referenced_table_name,
                lower(s2.rdb$field_name) as referenced_column_name
            from rdb$index_segments s
            left join rdb$indices i on i.rdb$index_name = s.rdb$index_name
            left join rdb$relation_constraints rc on rc.rdb$index_name = s.rdb$index_name
            left join rdb$ref_constraints refc on rc.rdb$constraint_name = refc.rdb$constraint_name
            left join rdb$relation_constraints rc2 on rc2.rdb$constraint_name = refc.rdb$const_name_uq
            left join rdb$indices i2 on i2.rdb$index_name = rc2.rdb$index_name
            left join rdb$index_segments s2 on i2.rdb$index_name = s2.rdb$index_name
            WHERE RC.RDB$CONSTRAINT_TYPE = 'FOREIGN KEY'
            and upper(i.rdb$relation_name) = %s """ % (tbl_name,))

        for r in cursor.fetchall():
            key_columns.append((r[0].strip(), r[1].strip(), r[2].strip()))
        return key_columns

    def get_relations(self, cursor, table_name):
        """
        Returns a dictionary of {field_name: (field_name_other_table, other_table)}
        representing all relationships to the given table.
        """
        constraints = self.get_key_columns(cursor, table_name)
        relations = {}
        for my_fieldname, other_table, other_field in constraints:
            relations[my_fieldname] = (other_field, other_table)
        return relations

    def get_constraints(self, cursor, table_name):
        """
        Retrieves any constraints or keys (unique, pk, fk, check, index)
        across one or more columns.

        Returns a dict mapping constraint names to their attributes,
        where attributes is a dict with keys:
         * columns: List of columns this covers
         * primary_key: True if primary key, False otherwise
         * unique: True if this is a unique constraint, False otherwise
         * foreign_key: (table, column) of target, or None
         * check: True if check constraint, False otherwise
         * index: True if index, False otherwise.

        Some backends may return special constraint names that don't exist
        if they don't name constraints of a certain type (e.g. SQLite)
        """
        # TODO: FIX check constraint
        tbl_name = "'%s'" % table_name.upper()
        constraints = {}

        cursor.execute("""
        SELECT
          case
            when rc.RDB$CONSTRAINT_NAME is not null then rc.RDB$CONSTRAINT_NAME
            else i.RDB$INDEX_NAME
          end as constraint_name,

          case
            when rc.RDB$CONSTRAINT_TYPE is not null then rc.RDB$CONSTRAINT_TYPE
            else 'INDEX'
          end AS constraint_type,

          s.RDB$FIELD_NAME AS field_name,
          i2.RDB$RELATION_NAME AS references_table,
          s2.RDB$FIELD_NAME AS references_field,
          i.RDB$UNIQUE_FLAG
        FROM RDB$INDEX_SEGMENTS s
        LEFT JOIN RDB$INDICES i ON i.RDB$INDEX_NAME = s.RDB$INDEX_NAME
        LEFT JOIN RDB$RELATION_CONSTRAINTS rc ON rc.RDB$INDEX_NAME = s.RDB$INDEX_NAME
        LEFT JOIN RDB$REF_CONSTRAINTS refc ON rc.RDB$CONSTRAINT_NAME = refc.RDB$CONSTRAINT_NAME
        LEFT JOIN RDB$RELATION_CONSTRAINTS rc2 ON rc2.RDB$CONSTRAINT_NAME = refc.RDB$CONST_NAME_UQ
        LEFT JOIN RDB$INDICES i2 ON i2.RDB$INDEX_NAME = rc2.RDB$INDEX_NAME
        LEFT JOIN RDB$INDEX_SEGMENTS s2 ON i2.RDB$INDEX_NAME = s2.RDB$INDEX_NAME
        WHERE i.RDB$RELATION_NAME = %s
        ORDER BY s.RDB$FIELD_POSITION
        """ % (tbl_name,))
        for constraint_name, constraint_type, column, other_table, other_column, unique in cursor.fetchall():
            primary_key = False
            foreign_key = None
            check = False
            index = False
            constraint = constraint_name.strip()
            constraint_type = constraint_type.strip()
            column = column.strip().lower()
            if other_table:
                other_table = other_table.strip().lower()
            if other_column:
                other_column = other_column.strip().lower()

            if constraint_type == 'PRIMARY KEY':
                primary_key = True
            elif constraint_type == 'UNIQUE':
                unique = True
            elif constraint_type == 'FOREIGN KEY':
                foreign_key = (other_table, other_column,)
            elif constraint_type == 'INDEX':
                index = True

            if constraint not in constraints:
                constraints[constraint] = {
                    "columns": [],
                    "primary_key": primary_key,
                    "unique": unique,
                    "foreign_key": foreign_key,
                    "check": check,
                    "index": index,
                    "type": 'idx' if index else '',
                }
            # Record the details
            constraints[constraint]['columns'].append(column)

        if not constraints:
            raise ValueError("'%s' is not found." % table_name)
        return constraints

    def _get_field_indexes(self, table_name, field_name):
        """
          Return a list of index (index_name, type, constraint_name)
        """
        with self.connection.cursor() as cursor:

            table = "'%s'" % table_name.upper()
            field = "'%s'" % field_name.upper()
            cursor.execute("""
                select
                    trim(s.rdb$index_name),
                    trim(rc.rdb$constraint_name),
                    trim(rc.rdb$constraint_type)
                from rdb$index_segments s
                left join rdb$indices i on i.rdb$index_name = s.rdb$index_name
                left join rdb$relation_constraints rc on rc.rdb$index_name = s.rdb$index_name
                where i.rdb$relation_name = %s
                and s.rdb$field_name = %s
                order by s.rdb$field_position """ % (table, field,))
    
            return [(i[0], i[1], i[2]) for i in cursor.fetchall()]



from django.db.backends.base.features import BaseDatabaseFeatures
from django.db.utils import InterfaceError


class DatabaseFeatures(BaseDatabaseFeatures):
    supports_forward_references = False
    supports_tablespaces = False
    has_zoneinfo_database = False
    uses_savepoints = True
    supports_paramstyle_pyformat = True
    has_native_uuid_field = False
    has_native_duration_field = False
    supports_column_check_constraints = False
    supports_regex_backreferencing = False
    has_bulk_insert = False
    can_return_id_from_insert = True
    has_select_for_update = False
    for_update_after_from = False
    supports_transactions = True
    can_introspect_small_integer_field = True
    supports_timezones = False
    closed_cursor_error_class = InterfaceError
    has_case_insensitive_like = False
    implied_column_null = True
    uppercases_column_names = True
    ignores_table_name_case = True
    truncates_names = True
    atomic_transactions = True
    supports_select_intersection = False
    supports_select_difference = False
    supports_index_column_ordering = False
    bare_select_suffix = " FROM RDB$DATABASE"
    requires_literal_defaults = True
    supports_cast_with_precision = False
    supports_sequence_reset = False
    supports_subqueries_in_group_by = False
    supports_partially_nullable_unique_constraints = False
    supports_mixed_date_datetime_comparisons = False
    can_introspect_autofield = True



import datetime
import uuid
import collections
from django.utils import timezone
from django.db.utils import InterfaceError

try:
    import firebirdsql as Database
except ImportError as e:
    raise ImproperlyConfigured("Error loading firebirdsql module: %s" % e)

def _quote_value(value):
    import binascii
    if isinstance(value, (datetime.date, datetime.time, datetime.datetime)):
        return "'%s'" % value
    elif isinstance(value, uuid.UUID):
        return "'%s'" % value.hex
    elif isinstance(value, str):
        return "'%s'" % value.replace("\'", "\'\'")
    elif isinstance(value, (bytes, bytearray, memoryview)):
        return "x'%s'" % binascii.hexlify(value).decode('ascii')
    elif value is None:
        return "NULL"
    else:
        return str(value)

def convert_sql(query, params):
    if params is None:
        pass
    elif isinstance(params, dict):
        converted_params = {}
        for k, v in params.items():
            if isinstance(v, datetime.datetime) and timezone.is_aware(v):
                v = v.astimezone(timezone.utc).replace(tzinfo=None)
            converted_params[k] = _quote_value(v)
        query = query % converted_params
    elif isinstance(params, (list, tuple)):
        converted_params = []
        for p in params:
            v = p
            if isinstance(v, datetime.datetime) and timezone.is_aware(v):
                v = v.astimezone(timezone.utc).replace(tzinfo=None)
            converted_params.append(_quote_value(v))
        if len(converted_params) == 1:
            query = query % converted_params[0]
        else:
            query = query % tuple(converted_params)
    return query


class FirebirdCursorWrapper(Database.Cursor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._rows = collections.deque()
        self.closed = False

    def execute(self, query, params=None):
        if self.closed:
            raise InterfaceError('Cursor is closed')
        super().execute(convert_sql(query, params))
        self._rows = collections.deque(super().fetchall())
        if self._transaction._autocommit:
            self._transaction._connection.commit()

    def executemany(self, query, param_list):
        if self.closed:
            raise InterfaceError('Cursor is closed')
        for params in param_list:
            super().execute(convert_sql(query, params))

    def fetchone(self):
        if len(self._rows):
            return self._rows.popleft()
        return None

    def fetchmany(self, size=1):
        rs = []
        for i in range(size):
            r = self.fetchone()
            if not r:
                break
            rs.append(r)
        return rs

    def fetchall(self):
        r = list(self._rows)
        self._rows.clear()
        return r

    def close(self):
        super().close()
        self.closed = True



import datetime

from django.db.models.fields import AutoField
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.backends.base.schema import _related_non_m2m_objects
from django.utils.encoding import force_text

from .cursor import FirebirdCursorWrapper, _quote_value     # NOQA isort:skip


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    sql_rename_table = "Rename table is not allowed"  # Not supported
    sql_delete_table = "DROP TABLE %(table)s"
    sql_create_column = "ALTER TABLE %(table)s ADD %(column)s %(definition)s"
    sql_alter_column_type = "ALTER %(column)s TYPE %(type)s"
    sql_alter_column_default = "ALTER COLUMN %(column)s SET DEFAULT %(default)s"
    sql_alter_column_no_default = "ALTER COLUMN %(column)s DROP DEFAULT"
    sql_delete_column = "ALTER TABLE %(table)s DROP %(column)s"
    sql_rename_column = "ALTER TABLE %(table)s ALTER %(old_column)s TO %(new_column)s"
    sql_create_fk = "ALTER TABLE %(table)s ADD CONSTRAINT %(name)s FOREIGN KEY (%(column)s) REFERENCES %(to_table)s (%(to_column)s) ON DELETE CASCADE"
    sql_delete_fk = "ALTER TABLE %(table)s DROP CONSTRAINT %(name)s"
    sql_pk_to_unique = "ALTER TABLE %(table)s DROP CONSTRAINT %(name)s, ADD CONSTRAINT %(name)s UNIQUE (%(column)s)"
    sql_unique_to_pk = "ALTER TABLE %(table)s DROP CONSTRAINT %(name)s, ADD CONSTRAINT %(name)s PRIMARY KEY (%(column)s)"
    sql_delete_constraint = "ALTER TABLE %(table)s DROP CONSTRAINT %(name)s"
    sql_add_identity = "ALTER TABLE %(table)s ALTER COLUMN %(column)s SET GENERATED BY DEFAULT"
    sql_delete_identity = "ALTER TABLE %(table)s ALTER COLUMN %(column)s DROP IDENTITY"

    def quote_value(self, value):
        return _quote_value(value)

    def prepare_default(self, value):
        return self.quote_value(value)

    def _get_field_indexes(self, model, field):
        return self.connection.introspection._get_field_indexes(model._meta.db_table, field.column)

    def _alter_column_type_sql(self, model, old_field, new_field, new_type):
        if new_field.get_internal_type() == 'AutoField':
            new_type = 'integer'
        elif new_field.get_internal_type() == 'BigAutoField':
            new_type = 'bigint'
        return super()._alter_column_type_sql(model, old_field, new_field, new_type)

    def _alter_field(self, model, old_field, new_field, old_type, new_type,
                     old_db_params, new_db_params, strict=False):
        if (old_field.get_internal_type() in ('AutoField', 'BigAutoField')
            and new_field.get_internal_type() not in ('AutoField', 'BigAutoField')):
            self.execute(self.sql_delete_identity % {
                'table': self.quote_name(model._meta.db_table),
                'column': self.quote_name(old_field.column),
            })

        if old_field.primary_key and new_field.unique:
            for index_name, constraint_name, constraint_type in self._get_field_indexes(model, old_field):
                if constraint_type == 'PRIMARY KEY':
                    self.execute(self.sql_pk_to_unique % {
                        'name': self.quote_name(constraint_name),
                        'table': self.quote_name(model._meta.db_table),
                        'column': self.quote_name(old_field.column),
                    })
        elif old_field.unique and new_field.primary_key:
            for index_name, constraint_name, constraint_type in self._get_field_indexes(model, old_field):
                if constraint_type == 'UNIQUE':
                    self.execute(self.sql_unique_to_pk % {
                        'name': self.quote_name(constraint_name),
                        'table': self.quote_name(model._meta.db_table),
                        'column': self.quote_name(old_field.column),
                    })

        super()._alter_field(model, old_field, new_field, old_type, new_type,
                     old_db_params, new_db_params)

        if (old_field.get_internal_type() not in ('AutoField', 'BigAutoField') and
            new_field.get_internal_type() in ('AutoField', 'BigAutoField')):
            self.execute(self.sql_add_identity % {
                'table': self.quote_name(model._meta.db_table),
                'column': self.quote_name(old_field.column),
            })

    def remove_field(self, model, field):
        # TODO:
        super().remove_field(model, field)

    def delete_model(self, model):
        """Delete a model from the database."""
        # delete related foreign key constraints
        for r in self.connection.introspection._get_references(model._meta.db_table):
            self.execute(self.sql_delete_fk % {'name': r[0], 'table': r[1].upper()})
        super().delete_model(model)

    def _column_has_default(self, params):
        sql = """
            SELECT a.RDB$DEFAULT_VALUE
            FROM RDB$RELATION_FIELDS a
            WHERE UPPER(a.RDB$FIELD_NAME) = UPPER('%(column)s')
            AND UPPER(a.RDB$RELATION_NAME) = UPPER('%(table_name)s')
        """
        value = self.execute(sql % params)
        return True if value else False

    def _column_sql(self, model, field):
        """
        Take a field and return its column definition.
        The field must already have had set_attributes_from_name() called.
        """
        # Get the column's type and use that as the basis of the SQL
        db_params = field.db_parameters(connection=self.connection)
        sql = db_params['type']

        # Primary key/unique outputs
        if field.primary_key:
            sql += " PRIMARY KEY"
        elif field.unique:
            sql += " UNIQUE"

        # Return the sql
        return sql



import sys
import firebirdsql as Database
from django.db.backends.base.creation import BaseDatabaseCreation

class DatabaseCreation(BaseDatabaseCreation):
    def _get_test_db_name(self):
        return self.connection.settings_dict['NAME']

    def _check_active_connection(self, verbosity):
        if self.connection:
            if verbosity >= 1:
                print("Closing active connection")
            self.connection.close()

    def _get_connection_params(self, **overrides):
        settings_dict = self.connection.settings_dict
        conn_params = {'charset': 'UTF8'}
        conn_params['database'] = settings_dict['NAME']
        if settings_dict['HOST']:
            conn_params['host'] = settings_dict['HOST']
        if settings_dict['PORT']:
            conn_params['port'] = settings_dict['PORT']
        if settings_dict['USER']:
            conn_params['user'] = settings_dict['USER']
        if settings_dict['PASSWORD']:
            conn_params['password'] = settings_dict['PASSWORD']
        if 'ROLE' in settings_dict:
            conn_params['role'] = settings_dict['ROLE']
        conn_params.update(settings_dict['OPTIONS'])
        conn_params.update(overrides)
        return conn_params

    def _get_creation_params(self, **overrides):
        settings_dict = self.connection.settings_dict
        params = {'charset': 'UTF8'}
        if settings_dict['USER']:
            params['user'] = settings_dict['USER']
        if settings_dict['PASSWORD']:
            params['password'] = settings_dict['PASSWORD']

        test_settings = settings_dict.get('TEST')
        if test_settings:
            if test_settings['NAME']:
                params['database'] = settings_dict['NAME']
            if test_settings['CHARSET']:
                params['charset'] = test_settings['CHARSET']
            if test_settings['PAGE_SIZE']:
                params['page_size'] = test_settings['PAGE_SIZE']
        params.update(overrides)
        return params

    def _create_database(self, test_database_name, verbosity):
        conn = Database.create_database(
                host=self.connection.settings_dict['HOST'],
                database=self.connection.settings_dict['NAME'],
                user=self.connection.settings_dict['USER'],
                password=self.connection.settings_dict['PASSWORD'],
                page_size=self.connection.settings_dict.get('PAGE_SIZE', 32768),
        )
        conn.close()

    def _create_test_db(self, verbosity, autoclobber, keepdb=False):
        """"
        Internal implementation - creates the test db tables.
        """
        test_database_name = self._get_test_db_name()
        self._create_database(test_database_name, verbosity)
        return test_database_name

    def _destroy_test_db(self, test_database_name, verbosity):
        """
        Internal implementation - remove the test db tables.
        """
        self._check_active_connection(verbosity)
        connection = Database.connect(**self._get_connection_params(database=test_database_name))
        connection.drop_database()
        connection.close()



import os
import subprocess

from django.db.backends.base.client import BaseDatabaseClient


class DatabaseClient(BaseDatabaseClient):

    @classmethod
    def runshell_db(cls, conn_params):
        if os.path.exists('/usr/bin/isql-fb'):
            executable_name = 'isql-fb'
        else:
            executable_name = 'isql'

        args = [executable_name]

        args.append(params['database'])

        if params['user']:
            args += ["-u", params['user']]
        if params['password']:
            args += ["-p", params['password']]
        if 'role' in params:
            args += ["-r", params['role']]

        subprocess.check_call(args)

    def runshell(self):
        DatabaseClient.runshell_db(self.connection.get_connection_params())



import uuid
import datetime
import pytz
import firebirdsql as Database

from django.conf import settings
from django.db.backends.utils import truncate_name
from django.db.backends.base.operations import BaseDatabaseOperations
from django.utils import timezone
from django.utils.encoding import force_text
from django.db.utils import DatabaseError
from django.db.models.functions import (
    ConcatPair, Substr, StrIndex, Repeat, Degrees, Radians
)


def _substr_as_sql(self, compiler, connection, function=None, template=None, arg_joiner=None, **extra_context):
    connection.ops.check_expression_support(self)
    sql_parts = []
    params = []
    for arg in self.source_expressions:
        arg_sql, arg_params = compiler.compile(arg)
        sql_parts.append(arg_sql)
        params.extend(arg_params)

    if len(sql_parts) == 2:
        template = 'SUBSTRING(%s FROM %s)' % (sql_parts[0], sql_parts[1])
    else:
        template = 'SUBSTRING(%s FROM %s FOR %s)' % (sql_parts[0], sql_parts[1], sql_parts[2])
    return template, params


def _str_index_as_sql(self, compiler, connection, function=None, template=None, arg_joiner=None, **extra_context):
    connection.ops.check_expression_support(self)
    sql_parts = []
    params = []
    for arg in self.source_expressions:
        arg_sql, arg_params = compiler.compile(arg)
        sql_parts.append(arg_sql)
        params.extend(arg_params)
    data = {**self.extra, **extra_context}
    if function is not None:
        data['function'] = function
    else:
        data.setdefault('function', self.function)
    template = 'POSITION(%(expressions)s)'
    arg_joiner = arg_joiner or data.get('arg_joiner', self.arg_joiner)
    sql_parts.reverse()
    data['expressions'] = data['field'] = arg_joiner.join(sql_parts)
    return template % data, params


ConcatPair.as_firebirdsql = ConcatPair.as_sqlite
Substr.as_firebirdsql = _substr_as_sql
StrIndex.as_firebirdsql = _str_index_as_sql
Repeat.as_firebirdsql = Repeat.as_oracle
Degrees.as_firebirdsql = Degrees.as_oracle
Radians.as_firebirdsql = Radians.as_oracle

class DatabaseOperations(BaseDatabaseOperations):
    cast_char_field_without_max_length = 'varchar(8191)'

    integer_field_ranges = {
        'SmallIntegerField': (-32768, 32767),
        'IntegerField': (-2147483648, 2147483647),
        'BigIntegerField': (-9223372036854775808, 9223372036854775807),
        'PositiveSmallIntegerField': (0, 32767),
        'PositiveIntegerField': (0, 2147483647),
    }

    cast_data_types = {
        'AutoField': 'integer',
        'BigAutoField': 'bigint',
    }

    def cache_key_culling_sql(self):
        return """
            SELECT cache_key
              FROM (SELECT cache_key, rank() OVER (ORDER BY cache_key) AS rank FROM %s)
             WHERE rank = %%s + 1
        """

    def unification_cast_sql(self, output_field):
        internal_type = output_field.get_internal_type()
        if internal_type in ("GenericIPAddressField", "IPAddressField", "TimeField", "UUIDField"):
            return 'CAST(%%s AS %s)' % output_field.db_type(self.connection).split('(')[0]
        return '%s'

    def check_expression_support(self, expression):
        from django.db.models.aggregates import Avg
        from django.db.models.expressions import Value
        from django.db.models.functions import (
            Greatest, Least, Length, Chr, LTrim, RTrim, Ord
        )

        if isinstance(expression, Avg):
            expression.template = '%(function)s(CAST(%(expressions)s as double precision))'
        elif isinstance(expression, Greatest):
            expression.function = 'MAXVALUE'
        elif isinstance(expression, Least):
            expression.function = 'MINVALUe'
        elif isinstance(expression, Length):
            expression.function = 'CHARACTER_LENGTH'
        elif isinstance(expression, Chr):
            expression.function = 'ASCII_CHAR'
        elif isinstance(expression, LTrim):
            expression.template = 'TRIM(LEADING FROM %(expressions)s)'
        elif isinstance(expression, RTrim):
            expression.template = 'TRIM(TRAILING FROM %(expressions)s)'
        elif isinstance(expression, Ord):
            expression.function = 'ASCII_VAL'
        elif isinstance(expression, Value):
            if isinstance(expression.value, datetime.datetime):
                expression.value = str(expression.value)[:24]

    def date_extract_sql(self, lookup_type, field_name):
        if lookup_type == 'week_day':
            return "EXTRACT(WEEKDAY FROM %s) + 1" % field_name
        elif lookup_type == 'quarter':
            return "((EXTRACT(MONTH FROM %s) - 1) / 3 + 1)" % field_name
        return "EXTRACT(%s FROM %s)" % (lookup_type.upper(), field_name)

    def date_interval_sql(self, timedelta):
        return timedelta

    def format_for_duration_arithmetic(self, sql):
        """Do nothing since formatting is handled in the custom function."""
        return sql

    def date_trunc_sql(self, lookup_type, field_name):
        if lookup_type == 'year':
            sql = "EXTRACT(year FROM %s)||'-01-01 00:00:00'" % field_name
        elif lookup_type == 'quarter':
            sql = "EXTRACT(year FROM %s)||'-'||((EXTRACT(MONTH FROM %s) -1) / 3 * 3 + 1)||'-01 00:00:00'" % (field_name, field_name)
        elif lookup_type == 'month':
            sql = "EXTRACT(year FROM %s)||'-'||EXTRACT(month FROM %s)||'-01 00:00:00'" % (field_name, field_name)
        elif lookup_type == 'week':
            sql = "DATEADD(day, EXTRACT(week FROM %s) * 7 - (EXTRACT(weekday FROM %s) + 1), CAST(EXTRACT(year FROM %s)||'-01-01 00:00:00' AS TIMESTAMP))" % (field_name, field_name, field_name)
        elif lookup_type == 'day':
            sql = "EXTRACT(year FROM %s)||'-'||EXTRACT(month FROM %s)||'-'||EXTRACT(day FROM %s)||' 00:00:00'" % (field_name, field_name, field_name)
        else:
            sql = field_name
        return "CAST(%s AS TIMESTAMP)" % sql

    def _convert_field_to_tz(self, field_name, tzname):
        if settings.USE_TZ:
            offset = int(datetime.datetime.now(pytz.timezone(tzname)).utcoffset().total_seconds())
            field_name = 'DATEADD(SECOND, %d, %s)' % (offset, field_name)
        return field_name

    def datetime_cast_date_sql(self, field_name, tzname):
        field_name = self._convert_field_to_tz(field_name, tzname)
        return 'CAST(%s AS DATE)' % field_name

    def datetime_cast_time_sql(self, field_name, tzname):
        field_name = self._convert_field_to_tz(field_name, tzname)
        return 'CAST(%s AS TIME)' % field_name

    def datetime_extract_sql(self, lookup_type, field_name, tzname):
        field_name = self._convert_field_to_tz(field_name, tzname)
        if lookup_type == 'week_day':
            sql = "EXTRACT(weekday FROM %s) + 1" % field_name
        elif lookup_type == 'quarter':
            sql = "((EXTRACT(month FROM %s) - 1) / 3 + 1)" % field_name
        else:
            sql = "EXTRACT(%s FROM %s)" % (lookup_type.upper(), field_name)
        return sql

    def datetime_trunc_sql(self, lookup_type, field_name, tzname):
        """
        Given a lookup_type of 'year', 'month', 'day', 'hour', 'minute' or
        'second', returns the SQL that truncates the given datetime field
        field_name to a datetime object with only the given specificity, and
        a tuple of parameters.
        """
        field_name = self._convert_field_to_tz(field_name, tzname)
        year = "EXTRACT(year FROM %s)" % field_name
        month = "EXTRACT(month FROM %s)" % field_name
        day = "EXTRACT(day FROM %s)" % field_name
        hh = "EXTRACT(hour FROM %s)" % field_name
        mm = "EXTRACT(minute FROM %s)" % field_name
        ss = "EXTRACT(second FROM %s)" % field_name
        yearday = "EXTRACT(yearday FROM %s)" % field_name
        weekday = "EXTRACT(weekday FROM %s)" % field_name
        quarter = "((EXTRACT(month FROM %s) -1) / 3 * 3 + 1)" % field_name
        if lookup_type == 'year':
            sql = "%s||'-01-01 00:00:00'" % year
        elif lookup_type == 'quarter':
            sql = "%s||'-'||%s||'-01 00:00:00'" % (year, quarter)
        elif lookup_type == 'month':
            sql = "%s||'-'||%s||'-01 00:00:00'" % (year, month)
        elif lookup_type == 'day':
            sql = "%s||'-'||%s||'-'||%s||' 00:00:00'" % (year, month, day)
        elif lookup_type == 'hour':
            sql = "%s||'-'||%s||'-'||%s||' '||%s||':00:00'" % (year, month, day, hh)
        elif lookup_type == 'minute':
            sql = "%s||'-'||%s||'-'||%s||' '||%s||':'||%s||':00'" % (year, month, day, hh, mm)
        elif lookup_type == 'second':
            sql = "%s||'-'||%s||'-'||%s||' '||%s||':'||%s||':'||%s" % (year, month, day, hh, mm, ss)
        elif lookup_type == 'week':
            sql = "DATEADD(day, IIF(%s = 0, -6, -%s+1), CAST(%s||'-'||%s||'-'||%s||' 00:00:00' AS TIMESTAMP))" % (weekday, weekday, year, month, day)
        return "CAST(%s AS TIMESTAMP)" % sql

    def time_trunc_sql(self, lookup_type, field_name):
        fields = {
            'hour': '%%H:00:00',
            'minute': '%%H:%%i:00',
            'second': '%%H:%%i:%%s',
        }  # Use double percents to escape.

        if lookup_type in fields:
            if lookup_type == 'hour':
                s = "EXTRACT(hour FROM %s)||':00:00'" % (field_name,)
            elif lookup_type == 'minute':
                s = "EXTRACT(hour FROM %s)||':'||EXTRACT(minute FROM %s)||':00'" % (field_name, field_name)
            elif lookup_type == 'second':
                s = "EXTRACT(hour FROM %s)||':'||EXTRACT(minute FROM %s)||':'||EXTRACT(second FROM %s)" % (field_name, field_name, field_name)
            return 'CAST(%s AS TIME)' % (s,)
        else:
            return "TIME(%s)" % (field_name)

    def no_limit_value(self):
        return None

    def limit_offset_sql(self, low_mark, high_mark):
        fetch, offset = self._get_limit_offset_params(low_mark, high_mark)
        return '%s%s' % (
            (' OFFSET %d ROWS' % offset) if offset else '',
            (' FETCH FIRST %d ROWS ONLY' % fetch) if fetch else '',
        )

    def quote_name(self, name):
        if not name.startswith('"') and not name.endswith('"'):
            name = '"%s"' % truncate_name(name, self.max_name_length())
        return name.upper()

    def sql_flush(self, style, tables, sequences, allow_cascade=False):
        if tables:
            sql = ['%s %s %s;' %
                    (style.SQL_KEYWORD('DELETE'),
                     style.SQL_KEYWORD('FROM'),
                     style.SQL_TABLE(self.quote_name(table))
                     ) for table in tables]
            return sql
        else:
            return []

    def sequence_reset_sql(self, style, model_list):
        from django.db import models
        output = []
        query = """EXECUTE BLOCK AS
            DECLARE S VARCHAR(255);
            DECLARE N INT;
            BEGIN
                SELECT MAX(%s) FROM %s INTO :N;
                EXECUTE STATEMENT 'ALTER TABLE %s ALTER COLUMN %s RESTART WITH ' || N;
            END"""
        for model in model_list:
            for f in model._meta.local_fields:
                if isinstance(f, models.AutoField):
                    output.append(query % (
                        self.quote_name(f.column),
                        self.quote_name(model._meta.db_table),
                        self.quote_name(model._meta.db_table),
                        self.quote_name(f.column),
                    ))
                    # Only one AutoField is allowed per model, so don't
                    # continue to loop
                    break
            for f in model._meta.many_to_many:
                if not f.remote_field.through:
                    output.append(query % (
                        self.quote_name('id'),
                        self.quote_name(f.m2m_db_table()),
                        self.quote_name(f.m2m_db_table()),
                        self.quote_name('id'),
                    ))
        return output

    def max_name_length(self):
        return 63

    def pk_default_value(self):
        return "DEFAULT"

    def return_insert_id(self):
        return "RETURNING %s", ()

    def random_function_sql(self):
        """
        Returns a SQL expression that returns a random value.
        """
        return 'RAND()'

    def prep_for_iexact_query(self, x):
        return x

    def adapt_datetimefield_value(self, value):
        """
        Transform a datetime value to an object compatible with what is expected
        by the backend driver for datetime columns.
        """
        if value is None:
            return None

        # Expression values are adapted by the database.
        if hasattr(value, 'resolve_expression'):
            return value

        # Firebird doesn't support tz-aware datetimes
        if timezone.is_aware(value):
            if settings.USE_TZ:
                value = timezone.make_naive(value, self.connection.timezone)
            else:
                raise ValueError("Firebird backend does not support timezone-aware datetimes when USE_TZ is False.")

        # Replaces 6 digits microseconds to 4 digits allowed in Firebird
        if isinstance(value, datetime.datetime):
            value = datetime.datetime(
                year=value.year,
                month=value.month,
                day=value.day,
                hour=value.hour,
                minute=value.minute,
                second=value.second,
                microsecond=(value.microsecond //100) * 100
            )
        return force_text(value)[:24]

    def adapt_timefield_value(self, value):
        if value is None:
            return None

        # Expression values are adapted by the database.
        if hasattr(value, 'resolve_expression'):
            return value

        # Firebird doesn't support tz-aware times
        if timezone.is_aware(value):
            raise ValueError("Firebird backend does not support timezone-aware times.")
        # Replaces 6 digits microseconds to 4 digits allowed in Firebird
        if isinstance(value, datetime.time):
            value = str(value)
        if isinstance(value, str):
            value = value[:13]
        return force_text(value)

    def combine_expression(self, connector, sub_expressions):
        lhs, rhs = sub_expressions
        if connector == '%%':
            return 'MOD(%s)' % ','.join(sub_expressions)
        elif connector == '&':
            return 'BIN_AND(%s)' % ','.join(sub_expressions)
        elif connector == '|':
            return 'BIN_AND(-%(lhs)s-1,%(rhs)s)+%(lhs)s' % {'lhs': lhs, 'rhs': rhs}
        elif connector == '<<':
            return '(%(lhs)s * POWER(2, %(rhs)s))' % {'lhs': lhs, 'rhs': rhs}
        elif connector == '>>':
            return 'FLOOR(%(lhs)s / POWER(2, %(rhs)s))' % {'lhs': lhs, 'rhs': rhs}
        elif connector == '^':
            return 'POWER(%s)' % ','.join(sub_expressions)
        return super().combine_expression(connector, sub_expressions)


    def get_db_converters(self, expression):
        converters = super().get_db_converters(expression)
        internal_type = expression.output_field.get_internal_type()
        if internal_type == 'DateTimeField':
            converters.append(self.convert_datetimefield_value)
        elif internal_type in ['IPAddressField', 'GenericIPAddressField']:
            converters.append(self.convert_ipfield_value)
        elif internal_type == 'UUIDField':
            converters.append(self.convert_uuidfield_value)
        return converters

    def convert_datetimefield_value(self, value, expression, connection):
        if value is not None:
            if settings.USE_TZ:
                value = timezone.make_aware(value, self.connection.timezone)
        return value

    def convert_ipfield_value(self, value, expression, connection):
        if value is not None:
            value = value.strip()
        return value

    def convert_uuidfield_value(self, value, expression, connection):
        if value is not None:
            value = uuid.UUID(value)
        return value

    def combine_duration_expression(self, connector, sub_expressions):
        if connector not in ['+', '-']:
            raise DatabaseError('Invalid connector for timedelta: %s.' % connector)
        sign = 1 if connector == '+' else -1

        sql, timedelta = sub_expressions
        if isinstance(sql, str) and isinstance(timedelta, str):
            if connector == '-':
                return 'DATEADD(MILLISECOND, -%s/1000, CAST(%s AS TIMESTAMP))' % (timedelta, sql)
            else:
                return 'DATEADD(MILLISECOND, %s/1000, CAST(%s AS TIMESTAMP))' % (timedelta, sql)

        if isinstance(sql, datetime.timedelta):
            sql, timedelta = timedelta, sql

        if isinstance(timedelta, str):
            return 'DATEADD(MILLISECOND, %s%s/1000, %s)' % (connector, timedelta, sql)
        elif timedelta.days:
            unit = 'day'
            value = timedelta.days
        elif timedelta.seconds:
            unit = 'second'
            value = ((timedelta.days * 86400) + timedelta.seconds)
        elif timedelta.microseconds:
            unit = 'millisecond'
            value = timedelta.microseconds // 1000
        else:
            unit = 'millisecond'
            value = 0
        return 'DATEADD(%s %s TO %s)' % (value * sign, unit, sql)

    def year_lookup_bounds_for_datetime_field(self, value):
        first = '%04d-01-01 00:00:00' % value
        second = '%04d-12-31 23:59:59.9999' % value
        return [first, second]

    def year_lookup_bounds_for_date_field(self, value):
        first = '%04d-01-01' % value
        second = '%04d-12-31' % value
        return [first, second]

    def lookup_cast(self, lookup_type, internal_type=None):
        if lookup_type in ('iexact', 'icontains', 'istartswith', 'iendswith'):
            return "UPPER(%s)"
        return "%s"

    def fulltext_search_sql(self, field_name):
        # We use varchar for TextFields so this is possible
        # Look at http://www.volny.cz/iprenosil/interbase/ip_ib_strings.htm
        return '%s CONTAINING %%s' % self.quote_name(field_name)

    def max_in_list_size(self):
        return 1500



"""
Firebird database backend for Django.

Requires firebirdsql: http://github.com/nakagami/pyfirebirdsql
"""
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends import utils

try:
    import firebirdsql as Database
except ImportError as e:
    raise ImproperlyConfigured("Error loading firebirdsql module: %s" % e)


from .client import DatabaseClient                          # NOQA isort:skip
from .creation import DatabaseCreation                      # NOQA isort:skip
from .features import DatabaseFeatures                      # NOQA isort:skip
from .introspection import DatabaseIntrospection            # NOQA isort:skip
from .operations import DatabaseOperations                  # NOQA isort:skip
from .schema import DatabaseSchemaEditor                    # NOQA isort:skip
from .cursor import FirebirdCursorWrapper, _quote_value     # NOQA isort:skip


class DatabaseWrapper(BaseDatabaseWrapper):
    vendor = 'firebirdsql'
    display_name = 'FirebirdSQL'
    # This dictionary maps Field objects to their associated FirebirdSQL column
    # types, as strings. Column-type strings can contain format strings; they'll
    # be interpolated against the values of Field.__dict__ before being output.
    # If a column type is set to None, it won't be included in the output.
    #
    # Any format strings starting with "qn_" are quoted before being used in the
    # output (the "qn_" prefix is stripped before the lookup is performed.
    data_types = {
        'AutoField': 'integer generated by default as identity',
        'BigAutoField': 'bigint generated by default as identity',
        'BinaryField': 'blob sub_type 0',
        'BooleanField': 'boolean',
        'CharField': 'varchar(%(max_length)s)',
        'DateField': 'date',
        'DateTimeField': 'timestamp',
        'DecimalField': 'decimal(%(max_digits)s, %(decimal_places)s)',
        'DurationField': 'bigint',
        'FileField': 'varchar(%(max_length)s)',
        'FilePathField': 'varchar(%(max_length)s)',
        'FloatField': 'double precision',
        'IntegerField': 'integer',
        'BigIntegerField': 'bigint',
        'IPAddressField': 'char(15)',
        'GenericIPAddressField': 'char(39)',
        'NullBooleanField': 'boolean',
        'OneToOneField': 'integer',
        'PositiveIntegerField': 'integer',
        'PositiveSmallIntegerField': 'smallint',
        'SlugField': 'varchar(%(max_length)s)',
        'SmallIntegerField': 'smallint',
        'TextField': 'blob sub_type 1',
        'TimeField': 'time',
        'UUIDField': 'char(32)',
    }
    data_type_check_constraints = {
        'PositiveIntegerField': '%(qn_column)s >= 0',
        'PositiveSmallIntegerField': '%(qn_column)s >= 0',
    }
    operators = {
        'exact': '= %s',
        'iexact': '= UPPER(%s)',
        'contains': "LIKE %s ESCAPE'\\'",
        'icontains': "LIKE UPPER(%s) ESCAPE'\\'",
        'regex': 'SIMILAR TO %s',
        'iregex': 'SIMILAR TO %s',
        'gt': '> %s',
        'gte': '>= %s',
        'lt': '< %s',
        'lte': '<= %s',
        'startswith': "LIKE %s ESCAPE'\\'",
        'endswith': "LIKE %s ESCAPE'\\'",
        'istartswith': "LIKE UPPER(%s) ESCAPE'\\'",
        'iendswith': "LIKE UPPER(%s) ESCAPE'\\'",
    }

    # The patterns below are used to generate SQL pattern lookup clauses when
    # the right-hand side of the lookup isn't a raw string (it might be an expression
    # or the result of a bilateral transformation).
    # In those cases, special characters for LIKE operators (e.g. \, *, _) should be
    # escaped on database side.
    #
    # Note: we use str.format() here for readability as '%' is used as a wildcard for
    # the LIKE operator.
    pattern_esc = r"REPLACE(REPLACE(REPLACE({}, '\', '\\'), '%%', '\%%'), '_', '\_')"
    pattern_ops = {
        'contains': "LIKE '%%' || {} || '%%' ESCAPE'\\'",
        'icontains': "LIKE '%%' || UPPER({}) || '%%' ESCAPE'\\'",
        'startswith': "LIKE {} || '%%' ESCAPE'\\'",
        'istartswith': "LIKE UPPER({}) || '%%' ESCAPE'\\'",
        'endswith': "LIKE '%%' || {} ESCAPE'\\'",
        'iendswith': "LIKE '%%' || UPPER({}) ESCAPE'\\'",
    }

    Database = Database
    SchemaEditorClass = DatabaseSchemaEditor
    # Classes instantiated in __init__().
    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_connection_params(self):
        settings_dict = self.settings_dict
        if not settings_dict['NAME']:
            raise ImproperlyConfigured(
                "settings.DATABASES is improperly configured. "
                "Please supply the NAME value.")
        conn_params = {'charset': 'UTF8'}
        conn_params['database'] = settings_dict['NAME']
        conn_params.update(settings_dict['OPTIONS'])
        if settings_dict['USER']:
            conn_params['user'] = settings_dict['USER']
        if settings_dict['PASSWORD']:
            conn_params['password'] = settings_dict['PASSWORD']
        if settings_dict['HOST']:
            conn_params['host'] = settings_dict['HOST']
        if settings_dict['PORT']:
            conn_params['port'] = settings_dict['PORT']
        return conn_params

    def get_new_connection(self, conn_params):
        connection = Database.connect(**conn_params)
        return connection

    def init_connection_state(self):
        self._set_autocommit(self.get_autocommit())

    def _set_autocommit(self, autocommit):
        with self.wrap_database_errors:
            self.connection.set_autocommit(autocommit)

    def create_cursor(self, name=None):
        return self.connection.cursor(factory=FirebirdCursorWrapper)

    def is_usable(self):
        return not self.connection.is_disconnect()

    def close_if_unusable_or_obsolete(self):
        if self.errors_occurred:
            self.close()
